#include <stdio.h>
#include <stdlib.h>

#include "image.h"
#include "mosaic.h"
#include <highgui.h>

static void PrintHelp();

// Print error message and exit
static void Fail(const char* s)
{
	printf("Error: %s\n", s);
	PrintHelp();
	exit(-1);
}

// Print help on how to use program
static void PrintHelp()
{
	printf("mosaic -o [output] [-l|-r|-u|-d] (-t###) (-w##) (-a) [input1] [input2] ...\n");
	printf("  output: Filename of output mosaic.\n");
	printf("  input:  Set of input files.  Must be listed after all.\n");
	printf("          other arguments.\n");
	printf("  -o:     Specify the output file. (default mosaic.bmp)\n");
	printf("  -l:     Camera moves left. (default)\n");
	printf("  -r:     Camera moves right.\n");
	printf("  -u:     Camera moves up.\n");
	printf("  -d:     Camera moves down.\n");
	printf("  -a:	  Use AVI file input.\n");
	printf("  -q:	  Use quick (translational) warping.\n");
	printf("  -t:###  Set convergence tolerance (default = .01).\n");
	printf("  -w:##   Set minimum pyramid dimension (default = 32).\n");

}

// Print help on how to use program
static char* outputFile = 0;
static char** inputFiles = 0;
static int numInputFiles = 0;
static char dir = 0;
static bool aviMode = false;
static bool use2 = false;
static float eps = .01f;
static unsigned int tip = 32;


const char D_LEFT = 'l';
const char D_RIGHT = 'r';
const char D_UP = 'u';
const char D_DOWN = 'd';


// Process command-line parameters
static void ProcessArgs(int argc, char ** argv)
{	
	aviMode = false;
	outputFile = 0;
	inputFiles = 0;
	numInputFiles = 0;
	dir = 0;

	for (int i = 1; i < argc; i++)
	{
		// Check that this is dashed argument
		if (argv[i][0] == '-')
		{
			// Ensure proper number of characters
			if (argv[i][1] == '\0')
				Fail("Invalid argument specified.\n");
				
			// Check dashed argument
			switch(argv[i][1])
			{
				case 'q':
					use2 = true;
				break;

				case 'a':
					aviMode = true;
				break;

				case 't':
					eps = max( min((float)atof(&argv[i][2]),1), .0001f);
				break;

				case 'w':
					tip = min( max(atoi(&argv[i][2]),16),256);
				break;

				// Output file
				case 'o':
				{
					// Ensure specified only once
					if (outputFile != 0)
						Fail("Already specified an output file.");

					// Check that file was specified
					if (i+1 >= argc || argv[i+1][0] == '-')
						Fail("No output file specified.");

					outputFile = argv[++i];
				} break;

				case D_LEFT:
				case D_RIGHT:
				case D_UP:
				case D_DOWN:
				{
					if (dir != 0)
						Fail("Specified two directions.");
					dir = argv[i][1];
					break;
				}
			}
		}
		else
		{
			inputFiles = &argv[i];
			numInputFiles = argc - i;
			// No dash, must be start of input files
			// Ensure no more parameters mixed in with input files
			for (; i < argc; i++)
			{
				if (argv[i][0] == '-')
					Fail("Must specify input files only after all other parameters.");
			}
		}
	}

	// Check that output file was specified
	if (!outputFile)
		outputFile = "mosaic.bmp";

	
	if (!dir)
		dir = D_LEFT;

	// Check that there are input files
	if ( aviMode && (numInputFiles != 1))
		Fail("Must specify one avi file as input.");
	else if (!aviMode && numInputFiles < 2)
		Fail("Must specify at least 2 input files.");
}

// Determine the starting arguments of the mosaic
int main(int argc, char ** argv)
{
	bool vertical, reverseDir;
	ProcessArgs(argc, argv);

	vertical = (dir == D_UP || dir == D_DOWN);
	reverseDir = (dir == D_LEFT || dir == D_UP);
  	
	printf("Using eps= %f, width= %d\n",eps,tip);
	
	IplImage* mosaic 
		= Mosaic(inputFiles, numInputFiles, vertical, reverseDir, aviMode,use2,eps,tip);

	if (mosaic)
	{
		if (dir == D_UP || dir == D_LEFT)
			MirrorIm(&mosaic, vertical);
		if(cvSaveImage(outputFile, (const CvArr*)mosaic))
			printf("Saved mosaic to: %s\n", outputFile);
		else
			printf("Failed to save mosaic to: %s\n", outputFile);
	}
	else
		printf("Could not create mosaic.\n");


	cvReleaseImage(&mosaic);
}
