#include "mosaic.h"
#include "LucasKanade.h"
#include "image.h"
#include "util.h"

#include <cv.h>
#include <float.h>
#include <highgui.h>
#include <math.h>
#include <vector>

static IplImage* AlignStrips(const IplImage * frame1, const IplImage * frame2,
							 int anchor, float* shift, int* extraX, int* dir,bool use2,float eps, unsigned int tip);

static IplImage* CreateWeightIm(int width, int height, int nChannels);

static IplImage* CloneBackwardFrame(IplImage* im);

// Construct a mosaic using the nIm files specified by imFile.
// If vertical is true, construct a vertical mosaic.  Otherwise
// construct a horizontal mosaic.
// If aviMode == true, construct a mosaic using the avi specified by
// imFile.
IplImage* Mosaic(char** imFile, int nIm, bool vertical, bool reverseDir, bool aviMode, bool use2, float eps, unsigned int tip)
{
	// Ensure valid set of files specified
	if (!aviMode && nIm < 2)
		return 0;

	// Setup arrays to store strip information
	std::vector<IplImage*> stripIm;
	std::vector<CvPoint2D32f> stripPos(1);
	stripIm.reserve(nIm);
	stripPos.reserve(nIm);
	stripPos[0] = cvPoint2D32f(0.0, 0.0);

	CvCapture * avi = NULL;

	// Load first frame
	IplImage *frame1;
	IplImage *capframe;
	if (aviMode)
	{
		avi = cvCaptureFromAVI(imFile[0]);
		if (!avi)
		{
			printf("Error loading AVI.\n");
			return NULL;
		}
		printf("Frame 0000\n");
		nIm = INT_MAX;
		frame1 = CloneUpsidedownImage(cvQueryFrame(avi));
	}
	else
	{
		frame1 = cvLoadImage(imFile[0],-1);
		printf("%s\n",imFile[0]);
	}
	if (vertical)
		TransposeIm(&frame1);

	// Ensure image has valid format
	if (!frame1 ||
		frame1->depth != IPL_DEPTH_8U ||
		frame1->nChannels != 3)
	{
		printf("Input must be 24-bit/32-bit color images\n");
		exit(-1);
	}

	IplImage *frame2;


	// Keep track of maximum and minimum y locations for strips
	float minY = 0.0, maxY = 0.0;
	int numStrips = 0;

	// Set anchor for strip alignment
	int anchor = frame1->width/2;

	int extraX = 0;

	// Obtain strips
	for (int i = 1; i < nIm; i++)
	{
		// Load next frame
		if (aviMode)
		{
			capframe = cvQueryFrame(avi);
			if (!capframe)
				break;

			frame2 = CloneUpsidedownImage(capframe);
			printf("Frame %04d",i);
		}
		else
		{
			frame2 = cvLoadImage(imFile[i],-1);
			printf("%s",imFile[i]);
		}
		if (vertical)
			TransposeIm(&frame2);

		// Ensure file has proper format
		if (!frame2 ||
			frame1->width != frame2->width ||
			frame1->height != frame2->height ||
			frame1->depth != frame2->depth ||
			frame1->nChannels != frame2->nChannels)
		{
			printf("\nError loading image: %s\n", imFile[i]);
			exit(-1);
		}

		// Align strips
		float shift;
		int dir;
		IplImage* im = AlignStrips(frame1, frame2, anchor, &shift, &extraX, &dir,use2,eps,tip);
		// If failed or going in wrong direction, skip this strip
		if (!im || (reverseDir && dir==1) || (!reverseDir && dir==-1)) {
			if (im)
				cvReleaseImage(&im);
			cvReleaseImage(&frame2);
			printf("-\n");
			continue;
		} else {
			printf("\n");
		}

		if (fabs(shift) > 50)
		{
			printf("large shift\n");
		}

		// Record strip
		stripIm.push_back(im);
		numStrips++;

		// Record position of strip
		stripPos[numStrips-1].x -= extraX;
		stripPos.push_back(cvPoint2D32f(
			stripPos[numStrips-1].x + stripIm[numStrips-1]->width - extraX,
			stripPos[numStrips-1].y + shift));

		// Update min-max y parameters
		if (stripPos[numStrips].y < minY)
			minY = stripPos[numStrips].y;

		if (stripPos[numStrips].y > maxY)
			maxY = stripPos[numStrips].y;

		// Reset frames
		cvReleaseImage(&frame1);
		frame1 = frame2;

	}
	
	// Construct mosaic sum and weight images
	CvSize sz;
	sz.width = (int)stripPos[numStrips].x;
	sz.height = (int)(ceil(maxY)-floor(minY) + frame1->height);

	if (sz.width == 0 || sz.height == 0) 
		return 0;

	IplImage *mosaicSum = cvCreateImage(sz, IPL_DEPTH_32F, frame1->nChannels);
	IplImage *mosaicWeight = cvCreateImage(sz, IPL_DEPTH_32F, frame1->nChannels);
	cvSetZero(mosaicSum);
	cvSet(mosaicWeight, cvScalar(FLT_EPSILON, FLT_EPSILON, FLT_EPSILON, FLT_EPSILON));

	// Copy strips into image
	for (int i = 0; i < numStrips-1; i++)
	{
		CvPoint2D32f pos = stripPos[i];
		pos.y -= minY;
		if (stripIm[i]->width == 0)
			continue;

		// Added weighted strip into mosaic images
		IplImage *stripMul 
			= CreateWeightIm(stripIm[i]->width, stripIm[i]->height, stripIm[i]->nChannels);
		WeightedBlit(stripIm[i], stripMul, mosaicSum, mosaicWeight, pos);

		// Release work images
		cvReleaseImage(&stripMul);
		cvReleaseImage(&stripIm[i]);
	}

	// Construct mosaic
	IplImage* mosaic = cvCreateImage(sz, frame1->depth, frame1->nChannels);
	IplImage* mosaicFloat = cvCreateImage(sz, IPL_DEPTH_32F, frame1->nChannels);
	cvDiv(mosaicSum, mosaicWeight, mosaicFloat);
	cvConvert(mosaicFloat, mosaic);

	if (aviMode)
		cvReleaseCapture(&avi);
	cvReleaseImage(&frame1);

	// Release work images
	cvReleaseImage(&mosaicFloat);
	cvReleaseImage(&mosaicWeight);
	cvReleaseImage(&mosaicSum);


	if (vertical)
		TransposeIm(&mosaic);
	return mosaic;
}


// Align frame1 and frame2 using the specified,  and return a strip 
// for the mosaic.
// Record the necessary vertical shift in shift, and the size of the
// horizontal borders outside the strip in extraX.
IplImage* AlignStrips(const IplImage * frame1, const IplImage * frame2,
					  int anchor, float* shift, int* extraX, int* dir,bool use2,float eps,unsigned int tip)
{
	CvMat * xform = cvCreateMat(2,3,CV_32F);	
	cvSetIdentity(xform);

	CvSize sz;
	sz.height = frame1->height;
	sz.width = frame1->width;

	// Register images
	if (!LKCoarseToFine(frame2,frame1,tip,xform,use2,eps))
		exit(-1);
	
	//PrintMatrix(xform);
	//ShowResults(frame2, frame1, xform);

	// Setup quad for warp 
	CvPoint2D32f pts[4];

	// Use anchor first and last points of quad
	pts[0].x = (float)anchor;  pts[0].y = (float)sz.height;
	pts[3].x = (float)anchor;  pts[3].y = 0;
	
	// Transform anchor of frame2 into frame1
	CvPoint2D32f anchorT[2];
	anchorT[0] = applyAffine(xform, pts[0]);
	anchorT[1] = applyAffine(xform, pts[3]);

	cvReleaseMat(&xform);

	// Move warped anchor so that its center is at the image center
	float center = sz.height / 2.0f;
	CvPoint2D32f diff = anchorT[0] - anchorT[1];
	float s = (center - (anchorT[0].y + anchorT[1].y)/2.0f) / diff.y;
	
	pts[1] = anchorT[0] + diff*s;
	pts[2] = anchorT[1] + diff*s;

	// Determine direction of mosaic
	if (pts[0].x <= pts[1].x && pts[3].x <= pts[2].x)
		*dir = 1;
	else if (pts[0].x >= pts[1].x && pts[3].x >= pts[2].x)
		*dir = -1;
	else
		*dir = 0;

	// Calculate strip shift
	*shift = length(anchorT[0] - pts[1]) * sz.height / length(pts[1] - pts[2]);
	if (s > 0)
		*shift *= -1;

	float len1 = length(pts[1] - pts[0]);
	float len2 = length(pts[3] - pts[2]);

	// Create strip
	CvSize stripSz;
	stripSz.height = frame1->height;
	stripSz.width = (int)(len1 > len2 ? len1 : len2);

	// Ensure we have a sane width and shift for the image
	float maxDist = sqrtf((float)frame1->width*frame1->width + (float)frame1->height*frame1->height);
	if (stripSz.width > maxDist || fabs(*shift) > maxDist)
	{
		return 0;
	}

	*extraX = stripSz.width;
	stripSz.width += *extraX*2;

	IplImage *strip = cvCreateImage(stripSz, frame1->depth, frame1->nChannels);

	// Warp pts in frame1 into strip
	QuadWarp(pts, *extraX, frame1, strip);

	return strip;
}


// Construct a weight image with specified width, height, and channels.
// The weight image is has a value of 1.0 at the center and linearly
// blends to 0 as pixels get to the left and right borders.
IplImage* CreateWeightIm(int width, int height, int nChannels)
{
	CvSize sz;
	sz.width = width;
	sz.height = height;
	IplImage *img = cvCreateImage(sz, IPL_DEPTH_32F, nChannels);

	int center = width/2;

	for (int x = 0; x < width; x++)
	{
		int dist = abs(x - center);
		float weight;
		if (center != 0)
			weight = 1.0f-(float)dist / center;
		else
			weight = 1.0;

		for (int y = 0; y < height; y++)
		{
			float* ptr = &((float*)(img->imageData + img->widthStep*y))[nChannels*x];
			for (int c = 0; c < nChannels; c++)
				ptr[c] = weight;
		}
	}

	return img;
}

